"""
STAC -> EO3 utilities
"""

from collections import namedtuple
from typing import Any, Dict, Iterable, List, Tuple, Iterator
from copy import deepcopy
from warnings import warn
import uuid

import toolz
from affine import Affine
import pystac.asset
import pystac.item
from pystac.extensions.projection import ProjectionExtension
from datacube.index.eo3 import prep_eo3
from datacube.index.index import default_metadata_type_docs
from datacube.model import Dataset, DatasetType, metadata_from_doc
from datacube.utils.geometry import GeoBox

BandMetadata = namedtuple("BandMetadata", ["dtype", "nodata", "units"])
ConversionConfig = Dict[str, Any]

(_eo3,) = [
    metadata_from_doc(d) for d in default_metadata_type_docs() if d.get("name") == "eo3"
]


def has_proj_ext(item: pystac.Item) -> bool:
    """
    Check if STAC Item has prjection extension
    """
    try:
        ProjectionExtension.validate_has_extension(item, add_if_missing=False)
        return True
    except pystac.ExtensionNotImplemented:
        return False


def has_proj_data(asset: pystac.asset.Asset) -> bool:
    """
    :returns: True if both ``.shape`` and ``.transform`` are set
    :returns: False if either ``.shape`` or ``.transform`` are missing
    """
    prj = ProjectionExtension.ext(asset)
    return prj.shape is not None and prj.transform is not None


def is_raster_data(asset: pystac.asset.Asset, check_proj: bool = False) -> bool:
    """
    - Has "data" role --> True
    - Has roles other than "data" --> False
    - Has no role but
      - media_type has ``image/``

    :param asset:
       STAC Asset to check

    :param check_proj:
       when enabled check if asset is part of an Item that has projection
       extension enabled and if yes only consider bands with
       projection data as "raster data" bands.
    """
    if check_proj:
        if has_proj_ext(asset.owner) and not has_proj_data(asset):
            return False

    if asset.roles is not None and len(asset.roles) > 0:
        return "data" in asset.roles
    return "image/" in asset.media_type


def asset_geobox(asset: pystac.asset.Asset) -> GeoBox:
    _proj = ProjectionExtension.ext(asset)
    assert _proj.shape is not None
    assert _proj.transform is not None
    assert _proj.crs_string is not None

    h, w = _proj.shape
    affine = Affine(*_proj.transform)
    return GeoBox(w, h, affine, _proj.crs_string)


def geobox_gsd(geobox: GeoBox) -> float:
    """
    Compute ground sampling distance of a given GeoBox.
    """
    return min(map(abs, [geobox.transform.a, geobox.transform.e]))


def compute_eo3_grids(
    assets: Dict[str, pystac.asset.Asset]
) -> Tuple[Dict[str, GeoBox], Dict[str, str]]:
    """
    Compute a minimal set of eo3 grids, pick default one, give names to
    non-default grids, while keeping track of which asset has which grid

    Assets must have ProjectionExtension with shape, transform and crs information
    populated.
    """

    def gbox_name(geobox: GeoBox) -> str:
        gsd = geobox_gsd(geobox)
        return f"g{gsd:g}"

    geoboxes = {k: asset_geobox(asset) for k, asset in assets.items()}

    # GeoBox to list of bands that share same footprint
    grids: Dict[GeoBox, List[str]] = {}

    for k, geobox in geoboxes.items():
        grids.setdefault(geobox, []).append(k)

    # Default grid is the one with highest count of bands
    #   If there is a tie pick one with the smallest ground sampling distance
    def gbox_score(geobox: GeoBox) -> Tuple[int, float]:
        return (-len(grids[geobox]), geobox_gsd(geobox))

    # locate default grid
    g_default, *_ = sorted(grids, key=gbox_score)

    named_grids: Dict[str, Geobox] = {}
    band2grid: Dict[str, str] = {}
    for grid, bands in grids.items():
        grid_name = "default" if grid is g_default else gbox_name(grid)
        if grid_name in named_grids:
            raise NotImplemented(
                "TODO: deal with multiple grids with same sampling distance"
            )

        named_grids[grid_name] = grid
        for band in bands:
            band2grid[band] = grid_name

    return named_grids, band2grid


def normalise_product_name(name: str) -> str:
    """
    Create valid product name from arbitrary string
    """

    # TODO: for now just map `-`,` ` to `_`
    return name.replace("-", "_").replace(" ", "_")


def mk_product(name: str, bands: Iterable[str], cfg: Dict[str, Any]) -> DatasetType:
    """
    Given a product name, list of bands and band metadata rules create EO3
    Datacube Product.
    """

    def _norm(meta) -> BandMetadata:
        if isinstance(meta, BandMetadata):
            return meta
        return BandMetadata(**meta)

    _cfg: Dict[str, BandMetadata] = {name: _norm(meta) for name, meta in cfg.items()}

    def make_band(name: str, cfg: Dict[str, BandMetadata]) -> Dict[str, Any]:
        info = cfg.get(name, cfg.get("*", BandMetadata("uint16", 0, "1")))

        return {
            "name": name,
            "dtype": info.dtype,
            "nodata": info.nodata,
            "units": info.units,
        }

    doc = {
        "name": normalise_product_name(name),
        "metadatat_type": "eo3",
        "measurements": [make_band(band, _cfg) for band in bands],
    }
    return DatasetType(_eo3, doc)


def infer_dc_product(item: pystac.Item, cfg: ConversionConfig) -> DatasetType:
    """
    :param item: Sample STAC Item from a collection
    :param cfg: Dictionary of configuration, see below

    .. code-block:: yaml

       sentinel-2-l2a:  # < name of the collection, i.e. ``.collection_id``
         measurements:
           "*":  # Band named "*" contains band info for "most" bands
             dtype: uint16
             nodata: 0
             units: "1"
           SCL:  # Those bands that are different than "most"
             dtype: uint8
             nodata: 0
             units: "1"
         uuid:   # Rules for constructing UUID for Datasets
             random:
             from_key: "location.of.unique.property"
             native: "location.of.key.with_actual_UUID"

       some-other-collection:
         measurements:
         #...
    """
    collection_id = item.collection_id

    cfg = deepcopy(cfg.get(collection_id, {}))

    data_bands = {
        name: asset
        for name, asset in item.assets.items()
        if is_raster_data(asset, check_proj=True)
    }

    product = mk_product(collection_id, data_bands, cfg.get("measurements", {}))

    # We assume that grouping of data bands into grids is consistent across
    # entire collection, so we compute it once and keep it on a product object
    # at least for now.
    _, band2grid = compute_eo3_grids(data_bands)
    cfg["band2grid"] = band2grid

    product._stac_cfg = cfg
    return product


def item_to_ds(item: pystac.Item, product: DatasetType) -> Dataset:
    _cfg = getattr(product, "_stac_cfg", {})

    _assets = item.assets
    data_bands = {name: _assets[name] for name in product.measurements}

    measurements: Dict[str, Dict[str, Any]] = {}
    grids: Dict[str, Dict[str, Any]] = {}
    crs = None

    for band, grid in _cfg["band2grid"].items():
        asset = _assets.get(band, None)
        if asset is None:
            warn(f"Missing asset with name: {band}")
            continue
        measurements[band] = {"path": asset.href}
        if grid != "default":
            measurements[band]["grid"] = grid

        if grid not in grids:
            geobox = asset_geobox(_assets[band])
            grids[grid] = dict(shape=geobox.shape, transform=geobox.transform)
            if crs is None:
                crs = geobox.crs
            else:
                if crs != geobox.crs:
                    raise ValueError("Expect all assets to share common CRS")

    assert crs is not None

    ds_uuid = str(uuid.uuid4())  # TODO: stop being so random

    ds_doc = {
        "id": ds_uuid,
        "$schema": "https://schemas.opendatacube.org/dataset",
        "crs": str(crs),
        "grids": grids,
        "location": "",
        "measurements": measurements,
        "properties": deepcopy(item.properties),
        "lineage": {},
    }

    return Dataset(product, prep_eo3(ds_doc), uris=[ds_doc.get("location", "")])


def stac2ds(items: Iterable[pystac.Item], cfg: ConversionConfig) -> Iterator[Dataset]:
    """
    Given a lazy sequence of STAC Items turn it into a lazy sequence of ``Dataset`` objects
    """
    products: Dict[str, DatasetType] = {}
    for item in items:
        product = products.get(item.collection_id)

        # Have not seen this collection yet, figure it out
        if product is None:
            product = infer_dc_product(item, cfg)
            products[item.collection_id] = product

        yield item_to_ds(item, product)



import math
from pathlib import Path
from typing import Any, Dict, Optional, Tuple
from uuid import UUID

from datacube.utils.geometry import Geometry
from odc.index import odc_uuid
from toolz import get_in

Document = Dict[str, Any]

# This is an old hack, should be refactored out
DEA_LANDSAT_PRODUCTS = ["ga_ls8c_ard_3", "ga_ls7e_ard_3", "ga_ls8t_ard_3"]

# This is an old hack too.
KNOWN_CONSTELLATIONS = ["sentinel-2"]

# Mapping between EO3 field names and STAC properties object field names
MAPPING_STAC_TO_EO3 = {
    "end_datetime": "dtr:end_datetime",
    "start_datetime": "dtr:start_datetime",
    "gsd": "eo:gsd",
    "instruments": "eo:instrument",
    "platform": "eo:platform",
    "constellation": "eo:constellation",
    "view:off_nadir": "eo:off_nadir",
    "view:azimuth": "eo:azimuth",
    "view:sun_azimuth": "eo:sun_azimuth",
    "view:sun_elevation": "eo:sun_elevation",
}

# Add more here as they are discovered
CANDIDATE_REGION_CODES = [
    "odc:region_code",
    "s2:mgrs_tile",
    "io:supercell_id"
]


def _get_region_code(properties: Dict[str, Any]) -> str:
    # Check the default candidates
    for region_code_name in CANDIDATE_REGION_CODES:
        region_code = properties.get(region_code_name, None)
        if region_code is not None:
            return region_code

    # Landsat special case
    if region_code is None:
        row = get_in(["landsat:wrs_row"], properties, None)
        path = get_in(["landsat:wrs_path"], properties, None)
        try:
            if row is not None and path is not None:
                region_code = f"{int(path):03d}{int(row):03d}"
        except ValueError:
            pass

    return region_code


def _stac_product_lookup(
    item: Document,
) -> Tuple[str, Optional[str], str, Optional[str], str]:
    properties = item["properties"]

    dataset_id: str = item["id"]
    dataset_label = item.get("title")

    product_name = properties.get("odc:product", None)
    if product_name is None:
        # If there's no odc:product or collection, then fail.
        product_name = get_in(["collection"], item, no_default=True)
    # Product names can't have dashes in them, for some reason
    product_name = product_name.replace("-", "_")

    # Try to get some region_codes
    region_code = _get_region_code(properties)

    default_grid = None

    # Maybe this should be the default product_name
    constellation = properties.get("constellation") or properties.get("eo:constellation")
    if constellation is not None:
        constellation = constellation.lower().replace(" ", "-")

    if constellation in KNOWN_CONSTELLATIONS:
        # This handles the Element 84 and  Microsft PC docs
        if constellation == "sentinel-2":
            dataset_id = properties.get("sentinel:product_id") or properties.get("s2:granule_id")
            product_name = "s2_l2a"
            if region_code is None:
                # Let this throw an exception if there's something missing
                region_code = "{}{}{}".format(
                    str(properties["proj:epsg"])[-2:],
                    properties["sentinel:latitude_band"],
                    properties["sentinel:grid_square"],
                )
            default_grid = "g10m"

    if product_name in DEA_LANDSAT_PRODUCTS:
        self_href = _find_self_href(item)
        dataset_label = Path(self_href).stem.replace(".stac-item", "")
        default_grid = "g30m"

    # If the ID is not cold and numerical, assume it can serve a label.
    if (
        not dataset_label
        and not _check_valid_uuid(dataset_id)
        and not dataset_id.isnumeric()
    ):
        dataset_label = dataset_id

    return dataset_id, dataset_label, product_name, region_code, default_grid


def _find_self_href(item: Document) -> str:
    """
    Extracting product label from filename of the STAC document 'self' URL
    """
    self_uri = [
        link.get("href", "")
        for link in item.get("links", [])
        if link.get("rel") == "self"
    ]

    if len(self_uri) < 1:
        raise ValueError("Can't find link for 'self'")
    if len(self_uri) > 1:
        raise ValueError("Too many links to 'self'")
    return self_uri[0]


def _get_stac_bands(
    item: Document,
    default_grid: str,
    relative: bool = False,
    proj_shape: Optional[str] = None,
    proj_transform: Optional[str] = None,
) -> Tuple[Document, Document, Document]:
    bands = {}
    grids = {}
    accessories = {}

    assets = item.get("assets", {})

    def _get_path(asset):
        path = asset["href"]
        if relative:
            path = Path(path).name

        return path

    for asset_name, asset in assets.items():
        # If something's not a geotiff, make it an accessory
        # include thumbnails in accessories
        if "geotiff" not in asset.get("type") or "thumbnail" in asset.get("roles", []):
            accessories[asset_name] = {"path": _get_path(asset)}
            continue

        # If transform specified here in the asset it should override
        # the properties-specified transform.
        transform = asset.get("proj:transform") or proj_transform
        grid = f"g{transform[0]:g}m"

        # As per transform, shape here overrides properties
        shape = asset.get("proj:shape") or proj_shape

        if grid not in grids:
            grids[grid] = {
                "shape": shape,
                "transform": transform,
            }

        path = _get_path(asset)
        band_index = asset.get("band", None)

        band_info = {"path": path}
        if band_index is not None:
            band_info["band"] = band_index

        # If we don't specify a default grid, label the first grid 'default'
        if not default_grid:
            default_grid = list(grids.keys())[0]

        if grid != default_grid:
            band_info["grid"] = grid

        bands[asset_name] = band_info

    if default_grid in grids:
        grids["default"] = grids.pop(default_grid)

    return bands, grids, accessories


def _geographic_to_projected(geometry, crs, precision=10):
    """Transform from WGS84 to the target projection, assuming Lon, Lat order"""
    geom = geometry.to_crs(crs, resolution=math.inf)

    def round_coords(c1, c2):
        return [round(coord, precision) for coord in [c1, c2]]

    if geom.is_valid:
        return geom.transform(round_coords)
    else:
        return None


def stac_transform_absolute(input_stac):
    return stac_transform(input_stac, relative=False)


def _convert_value_to_eo3_type(key: str, value):
    """
    Convert return type as per EO3 specification.
    Return type is String for "instrument" field in EO3 metadata.

    """
    if key == "instruments":
        if len(value) > 0:
            return "_".join([i.upper() for i in value])
        else:
            return None
    else:
        return value


def _get_stac_properties_lineage(input_stac: Document) -> Tuple[Document, Any]:
    """
    Extract properties and lineage field
    """
    properties = input_stac["properties"]
    prop = {
        MAPPING_STAC_TO_EO3.get(key, key): _convert_value_to_eo3_type(key, val)
        for key, val in properties.items()
    }

    creation_time = (
        properties.get("odc:processing_datetime")
        or properties.get("created")
        or properties.get("datetime")
    )
    if prop.get("odc:processing_datetime") is None and creation_time:
        prop["odc:processing_datetime"] = creation_time

    if prop.get("odc:file_format") is None:
        prop["odc:file_format"] = "GeoTIFF"

    # Extract lineage
    lineage = prop.pop("odc:lineage", None)

    return prop, lineage


def _check_valid_uuid(uuid_string: str) -> bool:
    """
    Check if provided uuid string is a valid UUID.
    """
    try:
        UUID(str(uuid_string))
        return True
    except ValueError:
        return False


def stac_transform(input_stac: Document, relative: bool = True) -> Document:
    """Takes in a raw STAC 1.0 dictionary and returns an ODC dictionary"""

    (
        dataset_id,
        dataset_label,
        product_name,
        region_code,
        default_grid,
    ) = _stac_product_lookup(input_stac)

    # Generating UUID for products not having UUID.
    # Checking if provided id is valid UUID.
    # If not valid, creating new deterministic uuid using odc_uuid function based on product_name and product_label.
    # TODO: Verify if this approach to create UUID is valid.
    if _check_valid_uuid(input_stac["id"]):
        deterministic_uuid = input_stac["id"]
    else:
        if product_name in ["s2_l2a"]:
            deterministic_uuid = str(
                odc_uuid("sentinel-2_stac_process", "1.0.0", [dataset_id])
            )
        else:
            deterministic_uuid = str(
                odc_uuid(f"{product_name}_stac_process", "1.0.0", [dataset_id])
            )

    # Check for projection extension properties that are not in the asset fields.
    # Specifically, proj:shape and proj:transform, as these are otherwise
    # fetched in _get_stac_bands.
    properties = input_stac["properties"]
    proj_shape = properties.get("proj:shape")
    proj_transform = properties.get("proj:transform")
    # TODO: handle old STAC that doesn't have grid information here...
    bands, grids, accessories = _get_stac_bands(
        input_stac,
        default_grid,
        relative=relative,
        proj_shape=proj_shape,
        proj_transform=proj_transform,
    )

    stac_properties, lineage = _get_stac_properties_lineage(input_stac)

    epsg = properties["proj:epsg"]
    native_crs = f"epsg:{epsg}"

    # Transform geometry to the native CRS at an appropriate precision
    geometry = Geometry(input_stac["geometry"], "epsg:4326")
    if native_crs != "epsg:4326":
        # Arbitrary precisions, but should be fine
        pixel_size = get_in(["default", "transform", 0], grids, no_default=True)
        precision = 0
        if pixel_size < 0:
            precision = 6

        geometry = _geographic_to_projected(geometry, native_crs, precision)

    stac_odc = {
        "$schema": "https://schemas.opendatacube.org/dataset",
        "id": deterministic_uuid,
        "crs": native_crs,
        "grids": grids,
        "product": {"name": product_name.lower()},
        "properties": stac_properties,
        "measurements": bands,
        "lineage": {},
        "accessories": accessories,
    }
    if dataset_label:
        stac_odc["label"] = dataset_label

    if region_code:
        stac_odc["properties"]["odc:region_code"] = region_code

    if geometry:
        stac_odc["geometry"] = geometry.json

    if lineage:
        stac_odc["lineage"] = lineage

    return stac_odc



"""
Test for SQS to DC tool
"""
import json
from pathlib import Path
import pytest
from datacube.utils import documents


TEST_DATA_FOLDER: Path = Path(__file__).parent.joinpath("data")
LANDSAT_STAC: str = "ga_ls8c_ard_3-1-0_088080_2020-05-25_final.stac-item.json"
LANDSAT_ODC: str = "ga_ls8c_ard_3-1-0_088080_2020-05-25_final.odc-metadata.yaml"
SENTINEL_STAC: str = "S2A_28QCH_20200714_0_L2A.json"
SENTINEL_STAC_MS: str = "S2B_MSIL2A_20190629T212529_R043_T06VVN_20201006T080531.json"
SENTINEL_ODC: str = "S2A_28QCH_20200714_0_L2A.odc-metadata.json"
USGS_LANDSAT_STAC: str = "LC08_L2SR_081119_20200101_20200823_02_T2.json"
LIDAR_STAC: str = "lidar_dem.json"

@pytest.fixture
def usgs_landsat_stac():
    with TEST_DATA_FOLDER.joinpath(USGS_LANDSAT_STAC).open("r") as f:
        return json.load(f)


@pytest.fixture
def landsat_stac():
    with TEST_DATA_FOLDER.joinpath(LANDSAT_STAC).open("r") as f:
        metadata = json.load(f)
    return metadata


@pytest.fixture
def lidar_stac():
    with TEST_DATA_FOLDER.joinpath(LIDAR_STAC).open("r") as f:
        metadata = json.load(f)
    return metadata


@pytest.fixture
def landsat_odc():
    metadata = yield from documents.load_documents(
        TEST_DATA_FOLDER.joinpath(LANDSAT_ODC)
    )
    return metadata


@pytest.fixture
def sentinel_stac():
    with TEST_DATA_FOLDER.joinpath(SENTINEL_STAC).open("r") as f:
        metadata = json.load(f)
    return metadata


@pytest.fixture
def sentinel_stac_ms():
    with TEST_DATA_FOLDER.joinpath(SENTINEL_STAC_MS).open("r") as f:
        metadata = json.load(f)
    return metadata


@pytest.fixture
def sentinel_odc():
    with TEST_DATA_FOLDER.joinpath(SENTINEL_ODC).open("r") as f:
        metadata = json.load(f)
    return metadata



"""
Test for stac_transform
"""
from functools import partial
from pprint import pformat
import pytest
from deepdiff import DeepDiff
from odc.index.stac import stac_transform


deep_diff = partial(
    DeepDiff, significant_digits=6, ignore_type_in_groups=[(tuple, list)]
)


def test_landsat_stac_transform(landsat_stac, landsat_odc):
    actual_doc = stac_transform(landsat_stac)
    do_diff(actual_doc, landsat_odc)


def test_sentinel_stac_transform(sentinel_stac, sentinel_odc):
    actual_doc = stac_transform(sentinel_stac)
    do_diff(actual_doc, sentinel_odc)


def test_usgs_landsat_stac_transform(usgs_landsat_stac):
    _ = stac_transform(usgs_landsat_stac)


def test_lidar_stac_transform(lidar_stac):
    _ = stac_transform(lidar_stac)


def do_diff(actual_doc, expected_doc):

    assert expected_doc["id"] == actual_doc["id"]
    assert expected_doc["crs"] == actual_doc["crs"]
    assert expected_doc["product"]["name"] == actual_doc["product"]["name"]
    assert expected_doc["label"] == actual_doc["label"]

    # Test geometry field
    doc_diff = deep_diff(expected_doc["geometry"], actual_doc["geometry"])
    assert doc_diff == {}, pformat(doc_diff)

    # Test grids field
    doc_diff = deep_diff(expected_doc["grids"], actual_doc["grids"])
    assert doc_diff == {}, pformat(doc_diff)

    # Test measurements field
    doc_diff = deep_diff(expected_doc["measurements"], actual_doc["measurements"])
    assert doc_diff == {}, pformat(doc_diff)

    # Test properties field
    doc_diff = deep_diff(
        expected_doc["properties"],
        actual_doc["properties"],
        exclude_paths=[
            "root['odc:product']",
            "root['proj:epsg']",
            "root['proj:shape']",
            "root['proj:transform']",
        ],
    )
    assert doc_diff == {}, pformat(doc_diff)

    # Test lineage field
    doc_diff = deep_diff(expected_doc["lineage"], actual_doc["lineage"])
    if expected_doc.get("accessories") is not None:
        doc_diff = deep_diff(expected_doc["accessories"], actual_doc["accessories"])
    assert doc_diff == {}, pformat(doc_diff)



import datetime
import jinja2


def render_eo3_yaml(task, processing_datetime=None, transform_precision=3):
    """

    task:
       .uuid          UUID
       .product       str
       .period        (datetime, datetime)
       .geobox        GeoBox
       .dss           [Dataset]
       .bands         [str]
       .region_code   str
       .file_prefix   str
    """
    if processing_datetime is None:
        processing_datetime = datetime.datetime.utcnow()

    return _YAML.render(
        task=task,
        transform_precision=transform_precision,
        processing_datetime=processing_datetime,
    )


_YAML = jinja2.Template(
    """---
# Dataset
$schema: https://schemas.opendatacube.org/dataset
id: {{ task.uuid}}

product:
  name: {{ task.product }}
  href: https://collections.dea.ga.gov.au/product/{{ task.product }}

crs: {{ task.geobox.crs }}
grids:
  default:
    shape: [{{ task.geobox.shape[0] }}, {{ task.geobox.shape[1] }}]
    transform: [{% for t in task.geobox.transform[:6] %}{{'%.*f' % (transform_precision, t)}},{% endfor %} 0, 0, 1]

properties:
  odc:region_code: {{ task.region_code }}
  datetime: {{ task.period[0].strftime("%Y-%m-%dT%H:%M:%S.%f") }}
  dtr:start_datetime: {{ task.period[0].strftime("%Y-%m-%dT%H:%M:%S.%f") }}
  dtr:end_datetime: {{ task.period[1].strftime("%Y-%m-%dT%H:%M:%S.%f") }}
  odc:file_format: GeoTIFF
  odc:processing_datetime: {{ processing_datetime.strftime("%Y-%m-%dT%H:%M:%S") }}

measurements:
{% for band in task.bands %}
  {{ band }}:
    path: {{ task.file_prefix }}_{{ band }}.tif
{% endfor %}

lineage:
  inputs:
{% for ds in task.dss %}
  - {{ ds.id }}
{% endfor %}
...
""",
    trim_blocks=True,
)



"""
STAC Item -> ODC Dataset[eo3]
"""
from ._version import __version__

from ._eo3 import (
    BandMetadata,
    ConversionConfig,
    stac2ds,
)


__all__ = (
    "BandMetadata",
    "ConversionConfig",
    "stac2ds",
    "__version__",
)



from odc.stac._eo3 import (
    mk_product,
    BandMetadata,
    compute_eo3_grids,
    infer_dc_product,
    is_raster_data,
    item_to_ds,
    stac2ds,
)
import pystac

STAC_CFG = {
    "sentinel-2-l2a": {
        "measurements": {
            "*": BandMetadata("uint16", 0, "1"),
            "SCL": BandMetadata("uint8", 0, "1"),
            "visual": dict(dtype="uint8", nodata=0, units="1"),
        }
    }
}


def test_mk_product():
    p = mk_product(
        "some-product",
        ["a", "b"],
        {"*": BandMetadata("uint8", 0, "1"), "b": BandMetadata("int16", -999, "BB")},
    )

    assert p.name == "some_product"
    assert p.metadata_type.name == "eo3"
    assert set(p.measurements) == set(["a", "b"])
    assert p.measurements["a"].dtype == "uint8"
    assert p.measurements["a"].nodata == 0
    assert p.measurements["a"].units == "1"

    assert p.measurements["b"].dtype == "int16"
    assert p.measurements["b"].nodata == -999
    assert p.measurements["b"].units == "BB"

    p = mk_product("Some Product", ["a", "b", "c"], {},)

    assert p.name == "Some_Product"
    assert set(p.measurements) == set(["a", "b", "c"])
    assert p.metadata_type.name == "eo3"

    for m in p.measurements.values():
        assert m.dtype == "uint16"
        assert m.nodata == 0
        assert m.units == "1"


def test_eo3_grids(sentinel_stac_ms):
    item = pystac.Item.from_dict(sentinel_stac_ms)
    assert item.collection_id == "sentinel-2-l2a"

    data_bands = {
        name: asset
        for name, asset in item.assets.items()
        if is_raster_data(asset, check_proj=True)
    }

    grids, b2g = compute_eo3_grids(data_bands)
    assert set(grids) == set("default g20 g60".split(" "))
    assert set(grids) == set(b2g.values())
    assert set(b2g) == set(data_bands)


def test_infer_product(sentinel_stac_ms):
    item = pystac.Item.from_dict(sentinel_stac_ms)

    assert item.collection_id in STAC_CFG

    product = infer_dc_product(item, STAC_CFG)
    assert product.measurements["SCL"].dtype == "uint8"
    assert product.measurements["visual"].dtype == "uint8"

    assert set(product._stac_cfg["band2grid"]) == set(product.measurements)


def test_item_to_ds(sentinel_stac_ms):
    item = pystac.Item.from_dict(sentinel_stac_ms)

    assert item.collection_id in STAC_CFG

    product = infer_dc_product(item, STAC_CFG)
    ds = item_to_ds(item, product)

    assert set(ds.measurements) == set(product.measurements)
    assert ds.crs is not None
    assert ds.metadata.lat is not None
    assert ds.metadata.lon is not None
    assert ds.center_time is not None

    dss = list(stac2ds(iter([item, item, item]), STAC_CFG))
    assert len(dss) == 3
    assert len(set(id(ds.type) for ds in dss)) == 1



""" Indexing related helper methods.
"""
from ..stac._version import __version__

from ._index import (
    from_metadata_stream,
    from_yaml_doc_stream,
    parse_doc_stream,
    bin_dataset_stream,
    bin_dataset_stream2,
    dataset_count,
    count_by_year,
    count_by_month,
    chop_query_by_time,
    time_range,
    month_range,
    season_range,
    ordered_dss,
    chopped_dss,
    all_datasets,
    product_from_yaml,
)

from ._uuid import (
    odc_uuid,
)

from ._utm import (
    utm_region_code,
    utm_zone_to_epsg,
    utm_tile_dss,
    mk_utm_gs,
)

from ._yaml import (
    render_eo3_yaml,
)

from ._grouper import (
    group_by_nothing,
    solar_offset,
)

__all__ = (
    "from_yaml_doc_stream",
    "from_metadata_stream",
    "parse_doc_stream",
    "bin_dataset_stream",
    "bin_dataset_stream2",
    "dataset_count",
    "count_by_year",
    "count_by_month",
    "chop_query_by_time",
    "time_range",
    "month_range",
    "season_range",
    "ordered_dss",
    "chopped_dss",
    "all_datasets",
    "product_from_yaml",
    "odc_uuid",
    "utm_region_code",
    "utm_zone_to_epsg",
    "utm_tile_dss",
    "mk_utm_gs",
    "render_eo3_yaml",
    "group_by_nothing",
    "solar_offset",
)



import sys
import datetime
import json
from warnings import warn
from types import SimpleNamespace
from typing import Iterator, Tuple, Optional

from datacube import Datacube
from datacube.api.query import Query
from datacube.index.hl import Doc2Dataset
from datacube.model import Range, Dataset, DatasetType, MetadataType, metadata_from_doc
from odc.io.text import parse_yaml
from ._grouper import solar_offset


def from_metadata_stream(metadata_stream, index, **kwargs):
    """
    Given a stream of (uri, metadata) tuples convert them into Datasets, using
    supplied index and options for Doc2Dataset.


    **kwargs**:
    - skip_lineage
    - verify_lineage
    - fail_on_missing_lineage
    - products
    - exclude_products

    returns a sequence of tuples where each tuple is either

        (Dataset, None) or (None, error_message)
    """
    doc2ds = Doc2Dataset(index, **kwargs)

    for uri, metadata in metadata_stream:
        if metadata is None:
            yield (None, "Error: empty doc %s" % (uri))
        else:
            ds, err = doc2ds(metadata, uri)
            if ds is not None:
                yield (ds, None)
            else:
                yield (None, "Error: %s, %s" % (uri, err))


def parse_doc_stream(doc_stream, on_error=None, transform=None):
    """Replace doc bytes/strings with parsed dicts.

       Stream[(uri, bytes)] -> Stream[(uri, dict)]


    :param doc_stream: sequence of (uri, doc: bytes|string)
    :param on_error: Callback uri, doc, exception -> None
    :param transform: dict -> dict if supplied also apply further transform on parsed document

    On output doc is replaced with python dict parsed from yaml, or with None
    if parsing/transform error occurred.
    """
    for uri, doc in doc_stream:
        try:
            if uri.endswith(".json"):
                metadata = json.loads(doc)
            else:
                metadata = parse_yaml(doc)

            if transform is not None:
                metadata = transform(metadata)
        except Exception as e:
            if on_error is not None:
                on_error(uri, doc, e)
            metadata = None

        yield uri, metadata


def from_yaml_doc_stream(doc_stream, index, logger=None, transform=None, **kwargs):
    """
    Stream[(path, bytes|str)] -> Stream[(Dataset, None)|(None, error_message)]

    :param doc_stream: sequence of (uri, doc: byges|string)
    :param on_error: Callback uri, doc, exception -> None
    :param logger:  Logger object for printing errors or None
    :param transform: dict -> dict if supplied also apply further transform on parsed document
    :param kwargs: passed on to from_metadata_stream

    """

    def on_parse_error(uri, doc, err):
        if logger is not None:
            logger.error(f"Failed to parse: {uri}")
        else:
            print(f"Failed to parse: {uri}", file=sys.stderr)

    metadata_stream = parse_doc_stream(
        doc_stream, on_error=on_parse_error, transform=transform
    )
    return from_metadata_stream(metadata_stream, index, **kwargs)


def dataset_count(index, **query):
    return index.datasets.count(**Query(**query).search_terms)


def count_by_year(index, product, min_year=None, max_year=None):
    """Returns dictionary Int->Int: `year` -> `dataset count for this year`.
    Only non-empty years are reported.
    """

    # TODO: get min/max from datacube properly
    if min_year is None:
        min_year = 1970
    if max_year is None:
        max_year = datetime.datetime.now().year

    ll = (
        (year, dataset_count(index, product=product, time=str(year)))
        for year in range(min_year, max_year + 1)
    )

    return {year: c for year, c in ll if c > 0}


def count_by_month(index, product, year):
    """Return 12 integer tuple
    counts for January, February ... December
    """
    return tuple(
        dataset_count(index, product=product, time="{}-{:02d}".format(year, month))
        for month in range(1, 12 + 1)
    )


def time_range(begin, end, freq="m"):
    """Return tuples of datetime objects aligned to boundaries of requested period
    (month is default).

    """
    from pandas import Period

    tzinfo = begin.tzinfo
    t = Period(begin, freq)

    def to_pydate(t):
        return t.to_pydatetime(warn=False).replace(tzinfo=tzinfo)

    while True:
        t0, t1 = map(to_pydate, (t.start_time, t.end_time))
        if t0 > end:
            break

        yield (max(t0, begin), min(t1, end))
        t += 1


def month_range(
    year: int, month: int, n: int
) -> Tuple[datetime.datetime, datetime.datetime]:
    """Return time range covering n months starting from year, month
    month 1..12
    month can also be negative
    2020, -1 === 2019, 12
    """
    if month < 0:
        return month_range(year - 1, 12 + month + 1, n)

    y2 = year
    m2 = month + n
    if m2 > 12:
        m2 -= 12
        y2 += 1
    dt_eps = datetime.timedelta(microseconds=1)

    return (
        datetime.datetime(year=year, month=month, day=1),
        datetime.datetime(year=y2, month=m2, day=1) - dt_eps,
    )


def season_range(year: int, season: str) -> Tuple[datetime.datetime, datetime.datetime]:
    """Season is one of djf, mam, jja, son.

    DJF for year X starts in Dec X-1 and ends in Feb X.
    """
    seasons = dict(djf=-1, mam=2, jja=6, son=9)

    start_month = seasons.get(season.lower())
    if start_month is None:
        raise ValueError(f"No such season {season}, valid seasons are: djf,mam,jja,son")
    return month_range(year, start_month, 3)


def chop_query_by_time(q: Query, freq: str = "m") -> Iterator[Query]:
    """Given a query over longer period of time, chop it up along the time dimension
    into smaller queries each covering a shorter time period (year, month, week or day).
    """
    qq = dict(**q.search_terms)
    time = qq.pop("time", None)
    if time is None:
        raise ValueError("Need time range in the query")

    for (t0, t1) in time_range(time.begin, time.end, freq=freq):
        yield Query(**qq, time=Range(t0, t1))


def ordered_dss(dc: Datacube, freq: str = "m", key=None, **query):
    """Emulate "order by time" streaming interface for datacube queries.

        Basic idea is to perform a lot of smaller queries (shorter time
        periods), sort results then yield them to the calling code.

    :param dc: Datacube instance

    :param freq: 'm' month sized chunks, 'w' week sized chunks, 'd' day

    :param key: Optional sorting function Dataset -> Comparable, for example
                ``lambda ds: (ds.center_time, ds.metadata.region_code)``
    """
    qq = Query(**query)
    if key is None:
        key = lambda ds: ds.center_time

    for q in chop_query_by_time(qq, freq=freq):
        dss = dc.find_datasets(**q.search_terms)
        dss.sort(key=key)
        yield from dss


def chopped_dss(dc: Datacube, freq: str = "m", **query):
    """Emulate streaming interface for datacube queries.

    Basic idea is to perform a lot of smaller queries (shorter time
    periods)
    """
    qq = Query(**query)

    for q in chop_query_by_time(qq, freq=freq):
        dss = dc.find_datasets_lazy(**q.search_terms)
        yield from dss


def bin_dataset_stream(gridspec, dss, cells, persist=None):
    """

    :param gridspec: GridSpec
    :param dss: Sequence of datasets (can be lazy)
    :param cells: Dictionary to populate with tiles
    :param persist: Dataset -> SomeThing mapping, defaults to keeping dataset id only

    The `cells` dictionary is a mapping from (x,y) tile index to object with the following properties

     .idx     - tile index (x,y)
     .geobox  - tile geobox
     .utc_offset - timedelta to add to timestamp to get day component in local time
     .dss     - list of UUIDs, or results of `persist(dataset)` if custom `persist` is supplied
    """

    geobox_cache = {}

    def default_persist(ds):
        return ds.id

    def register(tile, geobox, val):
        cell = cells.get(tile)
        if cell is None:
            utc_ofset = solar_offset(geobox.extent)
            cells[tile] = SimpleNamespace(
                geobox=geobox, idx=tile, utc_offset=utc_ofset, dss=[val]
            )
        else:
            cell.dss.append(val)

    if persist is None:
        persist = default_persist

    for ds in dss:
        ds_val = persist(ds)

        if ds.extent is None:
            warn("Dataset without extent info: %s" % str(ds.id))
            continue

        for tile, geobox in gridspec.tiles_from_geopolygon(
            ds.extent, geobox_cache=geobox_cache
        ):
            register(tile, geobox, ds_val)

        yield ds


def bin_dataset_stream2(gridspec, dss, geobox_cache=None):
    """
    For every input dataset compute tiles of the GridSpec it overlaps with.

    Iterable[Dataset] -> Iterator[(Dataset, List[Tuple[int, int]])]
    """
    if geobox_cache is None:
        geobox_cache = {}

    for ds in dss:
        if ds.extent is None:
            warn(f"Dataset without extent info: {ds.id}")
            tiles = []
        else:
            tiles = [
                tile
                for tile, _ in gridspec.tiles_from_geopolygon(
                    ds.extent, geobox_cache=geobox_cache
                )
            ]

        yield ds, tiles


def all_datasets(
    dc: Datacube, product: str, read_chunk: int = 1000, limit: Optional[int] = None
):
    """
    Like dc.find_datasets_lazy(product=product) but actually lazy, using db cursors
    """
    import psycopg2
    from random import randint

    assert isinstance(limit, (int, type(None)))

    db = psycopg2.connect(str(dc.index.url))
    _limit = "" if limit is None else f"LIMIT {limit}"

    _product = dc.index.products.get_by_name(product)

    query = f"""select
jsonb_build_object(
  'product', %(product)s,
  'uris', array((select _loc_.uri_scheme ||':'||_loc_.uri_body
                 from agdc.dataset_location as _loc_
                 where _loc_.dataset_ref = agdc.dataset.id and _loc_.archived is null
                 order by _loc_.added desc, _loc_.id desc)),
  'metadata', metadata) as dataset
from agdc.dataset
where archived is null
and dataset_type_ref = (select id from agdc.dataset_type where name = %(product)s)
{_limit};
"""
    cursor_name = "c{:04X}".format(randint(0, 0xFFFF))
    with db.cursor(name=cursor_name) as cursor:
        cursor.execute(query, dict(product=product))

        while True:
            chunk = cursor.fetchmany(read_chunk)
            if not chunk:
                break
            for (ds,) in chunk:
                yield Dataset(_product, ds["metadata"], ds["uris"])


def product_from_yaml(path: str, dc: Optional[Datacube] = None) -> DatasetType:
    """
    Make product definition from yaml file without access to the database.

    NOTE: access to database is only needed when non-standard metadata types
    are used.

    :param path: File path or a URL pointing to yaml definition of the product
    :param dc: Optional datacube instance (used to query MetadataType, only
               used if non-standard metadata type is used in the product)
    """
    from datacube.index.index import default_metadata_type_docs
    from datacube.utils.documents import load_documents

    standard_metadata_types = {
        d["name"]: metadata_from_doc(d) for d in default_metadata_type_docs()
    }

    product, *_ = load_documents(path)

    metadata = standard_metadata_types.get(product.get("metadata_type"), None)
    if metadata is not None:
        # Standard metadata we can do this without DB
        return DatasetType(metadata, product)

    # Not eo|eo3 see if we can get that from DB
    if dc is None:
        dc = Datacube()

    return dc.index.products.from_doc(product)
